import email.utils
import uuid
from collections import defaultdict
from datetime import datetime, timezone
from textwrap import dedent
from uuid import UUID

from django.conf import settings
from django.core.exceptions import ValidationError as DjangoValidationError
from django.db import connection, transaction
from django.db.models import CharField, Count, FloatField, Max, Prefetch, Q, QuerySet, Value, prefetch_related_objects
from django.db.models.functions import Cast
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from drf_spectacular.utils import (
    OpenApiParameter,
    PolymorphicProxySerializer,
    extend_schema,
    extend_schema_view,
    inline_serializer,
)
from psycopg2.extras import execute_values
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.mixins import DestroyModelMixin
from rest_framework.response import Response

from arkindex.documents.models import (
    AllowedMetaData,
    Classification,
    Corpus,
    Element,
    ElementPath,
    ElementType,
    MetaData,
    MetaType,
    MLClass,
    Selection,
    Transcription,
)
from arkindex.documents.serializers.elements import (
    CorpusSerializer,
    ElementBulkSerializer,
    ElementCreateSerializer,
    ElementDestinationSerializer,
    ElementListSerializer,
    ElementNeighborsSerializer,
    ElementParentSerializer,
    ElementSerializer,
    ElementSlimSerializer,
    ElementTypeSerializer,
    MetaDataBulkSerializer,
    MetaDataCreateSerializer,
    MetaDataSerializer,
    SelectionMoveSerializer,
    SelectionParentCreateSerializer,
)
from arkindex.documents.serializers.light import CorpusAllowedMetaDataSerializer, ElementTypeLightSerializer
from arkindex.documents.serializers.ml import ElementTranscriptionSerializer
from arkindex.images.models import Image
from arkindex.process.models import WorkerVersion
from arkindex.project.fields import Unnest
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, SelectionMixin
from arkindex.project.openapi import UUID_OR_FALSE, AutoSchema
from arkindex.project.pagination import LargePageNumberPagination, PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import BulkMap
from arkindex.project.triggers import (
    corpus_delete,
    element_trash,
    move_element,
    selection_worker_results_delete,
    worker_results_delete,
)
from arkindex.users.models import Role
from arkindex.users.utils import filter_rights

classifications_queryset = Classification.objects.select_related('ml_class', 'worker_version').order_by('-confidence')


def _fetch_has_children(elements):
    """
    From a list of elements, sets the `has_children` attribute to a boolean
    indicating if an element has children elements.

    The initial goal was to have a way for the frontend to tell if an element had any children;
    having a count was optional. It is possible to just use an Exists subquery since Django 2.2:
    Element.objects.annotate(has_children=Exists(ElementPath.objects.filter(path__last=OuterRef('pk'))))
    This works but makes the element list queries 20 times slower, due to the way the Postgres query planner
    handles this.

    We used a second query with a GROUP BY to retrieve the element path count, because the difference between
    telling if paths exists and getting a count is negligible with a GROUP BY, but the query could get very slow
    for elements with a lot of children, even without the count.

    We switched back to a boolean and to a query using a LEFT JOIN on a set of values, using the array overlap
    operator (&&) to match any child (not just direct children). This assumes that ElementPath is properly maintained
    and that no "ghost elements" exist in the paths.
    With an array instead of values, Postgres has to resort to a very slow sequential scan over the whole table.
    With direct children (the last element in a path), Postgres cannot use the GIN index on `path` and
    uses a sequential scan and a hash join.

    The ElementsListBase uses arkindex.project.tools.BulkMap to apply this method and perform the second request
    *after* DRF's pagination, because there is no way to perform post-processing after pagination in Django
    without having to use Django private methods.
    """
    # Ignore empty results
    if not elements:
        return elements

    with connection.cursor() as cursor:
        execute_values(
            cursor,
            'SELECT DISTINCT ON (e.id) e.id, p.id is not null as has_children '
            'FROM (VALUES %s) e (id) '
            'LEFT JOIN documents_elementpath p ON (ARRAY[e.id] && p.path)',
            tuple((element.id, ) for element in elements),
        )
        has_children = dict(cursor.fetchall())

    for element in elements:
        element.has_children = has_children[element.id]

    return elements


# Operators available for numeric filters in element lists
# Maps valid operator names in the API to Django QuerySet lookups
NUMERIC_OPERATORS = {
    'eq': 'exact',
    'lt': 'lt',
    'gt': 'gt',
    'lte': 'lte',
    'gte': 'gte',
}

# Operators available for metadata values.
METADATA_OPERATORS = {
    # Only for numeric metadata
    **NUMERIC_OPERATORS,
    # The contains operator should be case-insensitive
    'contains': 'icontains',
}

# Operators that do not require the metadata value to be a number
METADATA_STRING_OPERATORS = {'eq', 'contains'}

ELEMENT_ORDERING_FIELDS = {'name', 'created', 'random'}


class ElementsListAutoSchema(AutoSchema):
    """
    AutoSchema subclass that adds common parameters for all element list endpoints.
    This is not doable using @extend_schema as this causes the same schema to be reused for each endpoint
    subclassing ElementsListBase, causing duplicate operation IDs and invalid parameters.
    """

    def get_override_parameters(self):
        # Parameters that apply to all methods
        parameters = [
            OpenApiParameter(
                'type',
                description='Filter elements by type',
                required=False,
            ),
            OpenApiParameter(
                'name',
                description='Only include elements whose name contains a given string (case-insensitive)',
                required=False,
            ),
            OpenApiParameter(
                'folder',
                description='Restrict to or exclude elements with folder types',
                type=bool,
                required=False,
            ),
            OpenApiParameter(
                'worker_version',
                description='Only include elements created by a specific worker version. '
                            'If set to `False`, only include elements created by humans.',
                type=UUID_OR_FALSE,
                required=False,
            ),
            OpenApiParameter(
                'rotation_angle',
                description='Restrict to elements with the given rotation angle.',
                type={
                    'type': 'integer',
                    'minimum': 0,
                    'maximum': 359,
                },
                required=False,
            ),
            OpenApiParameter(
                'mirrored',
                description='Restrict to or exclude mirrored elements.',
                type=bool,
                required=False,
            )
        ]

        # Add method-specific parameters
        if self.method.lower() == 'get':
            parameters.extend([
                OpenApiParameter(
                    'confidence',
                    description='Restrict to elements with the given confidence. '
                                'The comparison operator can be set using `confidence_operator`.',
                    type={
                        'type': 'number',
                        'minimum': 0,
                        'maximum': 1
                    },
                    required=False,
                ),
                OpenApiParameter(
                    'confidence_operator',
                    description=dedent("""
                    Set the comparison operator to filter on element confidence scores:

                    * `eq` (default): Elements with this exact confidence.
                    * `lt`: Elements with a confidence strictly lower than the filter.
                    * `lte`: Elements with a confidence lower than or equal to the filter.
                    * `lt`: Elements with a confidence strictly greater than the filter.
                    * `gte`: Elements with a confidence greater than or equal to the filter.

                    This requires `confidence` to be set.
                    """),
                    enum=NUMERIC_OPERATORS.keys(),
                    default='eq',
                    required=False,
                ),
                OpenApiParameter(
                    'metadata_name',
                    description='Restrict to elements having a metadata with the given name.',
                    required=False,
                ),
                OpenApiParameter(
                    'metadata_value',
                    description='Restrict to elements having a metadata with the given value. '
                                'The comparison operator can be set using `metadata_operator`. '
                                'Requires `metadata_name` to be set.',
                    required=False,
                ),
                OpenApiParameter(
                    'metadata_operator',
                    description=dedent("""
                    Set the comparison operator to filter on metadata values:

                    * `eq` (default): Elements having a metadata with the given value.
                    * `contains`: Elements having a metadata whose value contains a given string.
                    * `lt`: Elements having a `numeric` metadata whose value is less than the filter.
                    * `lte`: Elements having a `numeric` metadata whose value is less than or equal to the filter.
                    * `gt`: Elements having a `numeric` metadata whose value is greater than the filter.
                    * `gte`: Elements having a `numeric` metadata whose value is greater than or equal to the filter.

                    This requires `metadata_name` and `metadata_value` to be set.
                    """),
                    enum=METADATA_OPERATORS.keys(),
                    default='eq',
                    required=False,
                ),
                OpenApiParameter(
                    'class_id',
                    description='Restrict to elements having a classification with the specified ML class ID.'
                                'If `classification_confidence` or `classification_high_confidence` are set, '
                                'the elements must have a classification that satisfies all of the parameters at once.',
                    type=UUID,
                    required=False,
                ),
                OpenApiParameter(
                    'classification_confidence',
                    description='Restrict to elements having a classification with the given confidence. '
                                'The comparison operator can be set using `classification_confidence_operator`. '
                                'If `class_id` or `classification_high_confidence` are set, the elements must have a '
                                'classification that satisfies all of the parameters at once.',
                    type={
                        'type': 'number',
                        'minimum': 0,
                        'maximum': 1
                    },
                    required=False,
                ),
                OpenApiParameter(
                    'classification_confidence_operator',
                    description=dedent("""
                    Set the comparison operator to filter on classification confidence scores:

                    * `eq` (default): Elements having a classification with this exact confidence.
                    * `lt`: Elements having a classification with a confidence strictly lower than the filter.
                    * `lte`: Elements having a classification with a confidence lower than or equal to the filter.
                    * `lt`: Elements having a classification with a confidence strictly greater than the filter.
                    * `gte`: Elements having a classification with a confidence greater than or equal to the filter.

                    This requires `classification_confidence` to be set.
                    """),
                    enum=NUMERIC_OPERATORS.keys(),
                    default='eq',
                    required=False,
                ),
                OpenApiParameter(
                    'classification_high_confidence',
                    description='Restrict to elements having a classification marked as `high_confidence`. '
                                'If `class_id` or `classification_confidence` are set, the elements must have a '
                                'classification that satisfies all of the parameters at once.',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'transcription_confidence',
                    description='Restrict to elements having a transcription with the given confidence. '
                                'The comparison operator can be set using `transcription_confidence_operator`. '
                                'If `transcription_worker_version` is set, the elements must have a '
                                'transcription that satisfies both filters at once.',
                    type={
                        'type': 'number',
                        'minimum': 0,
                        'maximum': 1
                    },
                    required=False,
                ),
                OpenApiParameter(
                    'transcription_confidence_operator',
                    description=dedent("""
                    Set the comparison operator to filter on transcription confidence scores:

                    * `eq` (default): Elements having a transcription with this exact confidence.
                    * `lt`: Elements having a transcription with a confidence strictly lower than the filter.
                    * `lte`: Elements having a transcription with a confidence lower than or equal to the filter.
                    * `lt`: Elements having a transcription with a confidence strictly greater than the filter.
                    * `gte`: Elements having a transcription with a confidence greater than or equal to the filter.

                    This requires `transcription_confidence` to be set.
                    """),
                    enum=NUMERIC_OPERATORS.keys(),
                    default='eq',
                    required=False,
                ),
                OpenApiParameter(
                    'transcription_worker_version',
                    description='Only include elements that have a transcription created by a specific worker version. '
                                'If set to `False`, only include elements with a transcription created by humans. '
                                'If `transcription_confidence` is set, the elements must have a '
                                'transcription that satisfies both filters at once.',
                    type=UUID_OR_FALSE,
                    required=False,
                ),
                OpenApiParameter(
                    'order',
                    description='Sort elements by a specific field, or sort randomly using `random`. '
                                'Random sorting is stateless: an element can be returned multiple times or not at all '
                                'in different pages of results, and requesting the same page twice will return different results.',
                    enum=ELEMENT_ORDERING_FIELDS,
                    default='name',
                    required=False,
                ),
                OpenApiParameter(
                    'order_direction',
                    description='Direction in which to sort elements. Ignored when using the `random` order.',
                    enum={'asc', 'desc'},
                    default='asc',
                    required=False,
                ),
                OpenApiParameter(
                    'with_classes',
                    description='Returns all classifications for each element. '
                                'Otherwise, `classes` will always be null.',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'with_metadata',
                    description='Returns all metadata without entities for each element. '
                                'Otherwise, `metadata` will always be null.',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'with_has_children',
                    description='Include the `has_children` boolean to tell if each element has direct children. '
                                'Otherwise, `has_children` will always be null.',
                    type=bool,
                    required=False,
                ),
                OpenApiParameter(
                    'with_corpus',
                    description='Returns the corpus attribute for each element. '
                                'When not set, the corpus attribute will be returned by default.',
                    type=bool,
                    default=True,
                    required=False,
                ),
                OpenApiParameter(
                    'with_zone',
                    description='Returns the zone attribute for each element. '
                                'When not set, the zone attribute will be returned by default.',
                    type=bool,
                    default=True,
                    required=False,
                ),
                OpenApiParameter(
                    'If-Modified-Since',
                    location=OpenApiParameter.HEADER,
                    description='Return HTTP 304 Not Modified when no elements have been created or updated since that time, '
                                'otherwise respond normally. Expects HTTP date formats (`Wed, 21 Oct 2015 07:28:00 GMT`). '
                                'Note that this is unable to detect deletions and will not filter the returned elements.',
                    required=False
                )
            ])

        elif self.method.lower() == 'delete':
            parameters.extend([
                OpenApiParameter(
                    'delete_children',
                    description='Delete all child elements of those elements recursively.',
                    type=bool,
                    required=False,
                )
            ])

        return parameters

    def get_tags(self):
        return ['elements']


class ElementsListBase(CorpusACLMixin, DestroyModelMixin, ListAPIView):
    """
    Provides common features across element list endpoints (list, parents, children)
    """
    schema = ElementsListAutoSchema()
    serializer_class = ElementListSerializer
    queryset = Element.objects.all()
    permission_classes = (IsVerifiedOrReadOnly, )
    pagination_class = LargePageNumberPagination

    def initial(self, *args, **kwargs):
        """
        Adds a `self.clean_params` attribute that holds any query params or headers,
        filtered by their allowed HTTP method.
        Does not perform validation on the filter values.
        """
        super().initial(*args, **kwargs)

        self.clean_params = {}
        # Set attributes for Spectacular's schema generation
        self.schema.path = ''
        self.schema.method = self.request.method
        for param in self.schema._get_parameters():
            if param['in'] == 'header':
                origin_dict = self.request.headers
            elif param['in'] == 'query':
                origin_dict = self.request.query_params
            else:
                raise NotImplementedError

            if param['name'] in origin_dict:
                self.clean_params[param['name']] = origin_dict[param['name']]

    def check_corpus_access(self, corpus):
        """
        Check that the corpus is accessible by the current user.
        This prevents comparing the corpus of every element.
        Require a contributor access for 'unsafe' methods (POST/PUT/PATCH).
        Require an admin access for deletion (DELETE).
        """
        role = Role.Contributor
        if self.request.method in permissions.SAFE_METHODS:
            role = Role.Guest
        elif self.request.method == 'DELETE':
            role = Role.Admin
        if not self.has_access(corpus, role.value):
            raise PermissionDenied(detail=f'You do not have {str(role).lower()} access to this corpus.')

    @cached_property
    def selected_corpus(self):
        # Retrieve the corpus and check user access rights only once
        corpus_id = self.kwargs.get('corpus')
        if corpus_id is None:
            return

        corpus = get_object_or_404(Corpus, id=corpus_id)
        self.check_corpus_access(corpus)
        return corpus

    @property
    def folder_filter(self):
        if self.clean_params.get('folder') is None:
            return
        return self.clean_params['folder'].lower() not in ('false', '0')

    @cached_property
    def type_filter(self):
        if 'type' not in self.request.query_params:
            return
        try:
            element_type = self.selected_corpus.types.get(slug=self.clean_params['type'])
        except ElementType.DoesNotExist:
            raise ValidationError({'type': ['This type does not exist.']})

        # When an API request is made with &type=…&folder=true, we avoid adding a type__folder filder in the query,
        # which would cause a join on ElementType, and instead check for the type's folder state here
        if self.folder_filter is not None and element_type.folder != self.folder_filter:
            raise ValidationError({'non_field_errors': [
                f'This type is {"not " if self.folder_filter else ""} a folder type.']
            })

        return element_type

    def get_worker_version_filter(self, filter_name):
        # If worker_version query param is False, we have to return elements created by humans
        if self.clean_params[filter_name].lower() in ('false', '0'):
            return None

        try:
            worker_version_id = uuid.UUID(self.clean_params[filter_name])
        except (TypeError, ValueError):
            raise ValidationError(['Invalid UUID'])

        if not WorkerVersion.objects.filter(id=worker_version_id).exists():
            raise ValidationError(['This worker version does not exist.'])

        return worker_version_id

    def get_metadata_queryset(self):
        """
        Returns a queryset that includes matched element IDs from metadata filters,
        or None if no metadata filters apply.
        """
        name = self.clean_params.get('metadata_name')
        value = self.clean_params.get('metadata_value')
        operator = self.clean_params.get('metadata_operator', '').lower().strip()
        if not any((name, value, operator)):
            return

        queryset = MetaData.objects.all()
        errors = defaultdict(list)

        if name:
            queryset = queryset.filter(name=name)

        if operator:
            if operator not in METADATA_OPERATORS:
                errors['metadata_operator'].append('This metadata operator is not supported.')
            if value is None:
                errors['metadata_operator'].append('This option is not supported without metadata_value.')
            elif operator not in METADATA_STRING_OPERATORS:
                try:
                    value = float(value)
                except (ValueError, TypeError):
                    errors['metadata_value'].append('Value must be a float when using a numeric comparison operator.')
                    value = None
        else:
            operator = 'eq'

        if value:
            lookup = METADATA_OPERATORS[operator]
            if name:
                if operator in METADATA_STRING_OPERATORS:
                    queryset = queryset.filter(**{f'value__{lookup}': value})
                else:
                    # Apply the numeric type filter, and filter on the value as a float
                    queryset = queryset \
                        .alias(numeric_value=Cast('value', output_field=FloatField())) \
                        .filter(type=MetaType.Numeric, **{f'numeric_value__{lookup}': value})
            else:
                errors['metadata_value'].append('This filter is not supported without metadata_name.')

        if errors:
            raise ValidationError(errors)

        return queryset.values('element_id')

    def get_classification_queryset(self):
        """
        Returns a queryset that includes matched element IDs from classification filters,
        or None if no classification filters apply.
        """
        class_id = self.clean_params.get('class_id')
        confidence = self.clean_params.get('classification_confidence')
        confidence_operator = self.clean_params.get('classification_confidence_operator', '').lower().strip()
        high_confidence = self.clean_params.get('classification_high_confidence', '').lower().strip()

        if len(high_confidence):
            high_confidence = high_confidence not in ('false', '0')
        else:
            # An empty string should be treated as no filter at all
            high_confidence = None

        if not class_id and confidence is None and not confidence_operator and high_confidence is None:
            # No filters apply
            return None

        queryset = Classification.objects.all()
        errors = defaultdict(list)

        if class_id:
            try:
                ml_class = self.selected_corpus.ml_classes.get(id=class_id)
            except DjangoValidationError as e:
                # An invalid UUID would cause a Django ValidationError
                errors['class_id'].extend(e.messages)
            except MLClass.DoesNotExist:
                errors['class_id'].append(f'ML class "{class_id}" not found')
            else:
                queryset = queryset.filter(ml_class=ml_class)

        if confidence_operator:
            if confidence_operator not in NUMERIC_OPERATORS:
                errors['classification_confidence_operator'].append('This operator is not supported.')
            if confidence is None:
                errors['classification_confidence_operator'].append('This option is not supported without classification_confidence.')
        else:
            confidence_operator = 'eq'

        if confidence:
            try:
                confidence = float(confidence)
                assert 0 <= confidence <= 1, 'Confidence must be between 0 and 1'
            except (TypeError, ValueError, AssertionError) as e:
                errors['classification_confidence'].append(str(e))
            else:
                lookup = NUMERIC_OPERATORS.get(confidence_operator, 'exact')
                queryset = queryset.filter(**{f'confidence__{lookup}': confidence})

        if high_confidence is not None:
            queryset = queryset.filter(high_confidence=high_confidence)

        if errors:
            raise ValidationError(errors)

        return queryset.values('element_id')

    def get_transcription_queryset(self):
        """
        Returns a queryset that includes matched element IDs from transcription filters,
        or None if no transcription filters apply.
        """
        worker_version_id = self.clean_params.get('transcription_worker_version')
        confidence = self.clean_params.get('transcription_confidence')
        confidence_operator = self.clean_params.get('transcription_confidence_operator', '').lower().strip()

        if not worker_version_id and confidence is None and not confidence_operator:
            # No filters apply
            return None

        queryset = Transcription.objects.all()
        errors = defaultdict(list)

        if worker_version_id:
            try:
                queryset = queryset.filter(worker_version_id=self.get_worker_version_filter('transcription_worker_version'))
            except ValidationError as e:
                errors['transcription_worker_version'] = e.detail

        if confidence_operator:
            if confidence_operator not in NUMERIC_OPERATORS:
                errors['transcription_confidence_operator'].append('This operator is not supported.')
            if confidence is None:
                errors['transcription_confidence_operator'].append('This option is not supported without transcription_confidence.')
        else:
            confidence_operator = 'eq'

        if confidence:
            try:
                confidence = float(confidence)
                assert 0 <= confidence <= 1, 'Confidence must be between 0 and 1'
            except (TypeError, ValueError, AssertionError) as e:
                errors['transcription_confidence'].append(str(e))
            else:
                lookup = NUMERIC_OPERATORS.get(confidence_operator, 'exact')
                queryset = queryset.filter(**{f'confidence__{lookup}': confidence})

        if errors:
            raise ValidationError(errors)

        return queryset.values('element_id')

    def get_filters(self) -> Q:
        filters = Q(corpus=self.selected_corpus)
        errors = defaultdict(list)

        if 'name' in self.clean_params:
            filters &= Q(name__icontains=self.clean_params['name'])

        if 'rotation_angle' in self.clean_params:
            try:
                rotation_angle = int(self.clean_params['rotation_angle'])
                assert 0 <= rotation_angle <= 359, 'A rotation angle must be between 0 and 359 degrees'
            except (AssertionError, TypeError, ValueError) as e:
                errors['rotation_angle'].append(str(e))
            else:
                filters &= Q(rotation_angle=rotation_angle)

        if 'mirrored' in self.clean_params:
            filters &= Q(mirrored=self.clean_params['mirrored'].lower() not in ('false', '0'))

        if self.type_filter:
            filters &= Q(type=self.type_filter)
        elif self.folder_filter is not None:
            # When filtering for folder or non-folder elements, using only the type__folder filter
            # can cause Postgres to retrieve all the {non-}folder types on every corpus
            # This can reach hundreds of types as the database grows, so Postgres can end up using a Hash Join
            # to handle joining this large amount of elements and types.
            # Since Postgres estimates this to represent a large amount of rows, it might also use multi-processing,
            # which has a very high overhead.
            # This can be avoided by also filtering on the type's corpus: Postgres will then access the index
            # on the type's corpus ID.  The query planner's statistics will give it a very low estimation since there
            # rarely are a ton of types in a corpus, so Postgres will also use the type_id index on elements, which
            # will lower the amount of rows much more quickly, making it stop using multi-processing.
            filters &= Q(type__corpus=self.selected_corpus, type__folder=self.folder_filter)

        if 'worker_version' in self.clean_params:
            try:
                filters &= Q(worker_version_id=self.get_worker_version_filter('worker_version'))
            except ValidationError as e:
                errors['worker_version'] = e.detail

        confidence = self.clean_params.get('confidence')
        confidence_operator = self.clean_params.get('confidence_operator', '').lower().strip()
        if confidence_operator:
            if confidence_operator not in NUMERIC_OPERATORS:
                errors['confidence_operator'].append('This operator is not supported.')
            if confidence is None:
                errors['confidence_operator'].append('This option is not supported without `confidence`.')
        else:
            confidence_operator = 'eq'

        if confidence:
            try:
                confidence = float(confidence)
                assert 0 <= confidence <= 1, 'Confidence must be between 0 and 1'
            except (TypeError, ValueError, AssertionError) as e:
                errors['confidence'].append(str(e))
            else:
                lookup = NUMERIC_OPERATORS.get(confidence_operator, 'exact')
                filters &= Q(**{f'confidence__{lookup}': confidence})

        try:
            metadata_queryset = self.get_metadata_queryset()
        except ValidationError as e:
            errors.update(e.detail)
        else:
            if metadata_queryset is not None:
                filters &= Q(id__in=metadata_queryset)

        try:
            classification_queryset = self.get_classification_queryset()
        except ValidationError as e:
            errors.update(e.detail)
        else:
            if classification_queryset is not None:
                filters &= Q(id__in=classification_queryset)

        try:
            transcription_queryset = self.get_transcription_queryset()
        except ValidationError as e:
            errors.update(e.detail)
        else:
            if transcription_queryset is not None:
                filters &= Q(id__in=transcription_queryset)

        if errors:
            raise ValidationError(errors)

        return filters

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus'] = self.selected_corpus
        if self.request:
            context['with_zone'] = self.request.query_params.get('with_zone', 'true').lower() not in ('false', '0')
            context['with_corpus'] = self.request.query_params.get('with_corpus', 'true').lower() not in ('false', '0')
        return context

    def get_prefetch(self):
        # Select the image and server in one separate query
        prefetch = {Prefetch('image', queryset=Image.objects.select_related('server')), 'type'}

        with_classes = self.clean_params.get('with_classes')
        if with_classes and with_classes.lower() not in ('false', '0'):
            prefetch.add(Prefetch('classifications', queryset=classifications_queryset, to_attr='classes'))

        with_metadata = self.clean_params.get('with_metadata')
        if with_metadata and with_metadata.lower() not in ('false', '0'):
            prefetch.add(Prefetch(
                'metadatas',
                queryset=MetaData.objects.filter(entity__isnull=True),
                to_attr='prefetched_metadata',
            ))

        return prefetch

    def get_order_direction(self):
        direction = self.clean_params.get('order_direction', 'asc').lower()
        if direction not in ('asc', 'desc'):
            raise ValidationError({'order_direction': ['Unknown sorting direction']})
        return '-' if direction == 'desc' else ''

    def get_order_by(self):
        sort_field = self.clean_params.get('order', 'name').lower()

        if sort_field not in ELEMENT_ORDERING_FIELDS:
            raise ValidationError({'order': ['Unknown sorting field']})

        if sort_field == 'random':
            return ('?', )

        direction = self.get_order_direction()
        # The sorting field is not unique; fallback on ordering by ID to ensure a consistent ordering
        return (f'{direction}{sort_field}', f'{direction}id')

    def filter_queryset(self, queryset):
        queryset = queryset \
            .filter(self.get_filters()) \
            .prefetch_related(*self.get_prefetch()) \
            .order_by(*self.get_order_by())

        with_has_children = self.clean_params.get('with_has_children')
        if with_has_children and with_has_children.lower() not in ('false', '0'):
            queryset = BulkMap(_fetch_has_children, queryset)

        return queryset

    def check_not_modified(self, queryset):
        """
        Returns True if HTTP 304 Not Modified should be returned. Requires a Django QuerySet instance or a BulkMap.
        """
        # Retrieve the inner QuerySet from the BulkMap if it is used
        if isinstance(queryset, BulkMap):
            queryset = queryset.iterable
        assert isinstance(queryset, QuerySet), 'A Django QuerySet is required to check for modified elements'

        modified_since_string = self.clean_params.get('If-Modified-Since')
        if not modified_since_string:
            return False

        # Parses RFC 2822 dates into tuples
        # Returns None if no date could be parsed or if the header was not defined
        time_tuple = email.utils.parsedate_tz(modified_since_string)
        if not time_tuple:
            raise ValidationError({'If-Modified-Since': ['Bad date format']})

        # Build a datetime from the tuple
        modified_since = datetime.fromtimestamp(email.utils.mktime_tz(time_tuple), timezone.utc)

        # Cause a HTTP 304 only if no element has been created or modified since then
        return not queryset.filter(Q(created__gt=modified_since) | Q(updated__gt=modified_since)).exists()

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        if self.check_not_modified(queryset):
            return Response(None, status=status.HTTP_304_NOT_MODIFIED)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def delete(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        delete_children = self.clean_params.get('delete_children', '').lower() not in ('false', '0')

        if not queryset.exists():
            raise NotFound

        element_trash(queryset, delete_children=delete_children, user_id=self.request.user.id)
        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(
    parameters=[
        OpenApiParameter(
            'top_level',
            description='Restrict to or exclude folder elements without parent elements (top-level elements).',
            type=bool,
            required=False,
        )
    ]
)
@extend_schema_view(
    get=extend_schema(operation_id='ListElements'),
    delete=extend_schema(operation_id='DestroyElements', description=(
        'Destroy elements in bulk.\n\n'
        'Requires an **admin** access to the corpus.'
    )),
)
class CorpusElements(ElementsListBase):
    """
    List elements in a corpus.

    Requires a **read** access to the corpus.
    """

    @property
    def is_top_level(self):
        if self.clean_params.get('top_level') is None:
            return None
        return self.clean_params['top_level'].lower() not in ('false', '0')

    def get_queryset(self):
        # Should not be possible due to the URL
        assert self.selected_corpus, 'Missing corpus ID'
        return self.selected_corpus.elements.all()

    def get_filters(self) -> Q:
        filters = super().get_filters()

        if self.is_top_level is not None:
            # Only include elements with an empty path.
            # An empty path's last item will be NULL, so we can use the index on path__last too
            filters &= Q(paths__path__last__isnull=self.is_top_level)

        return filters


@extend_schema(
    parameters=[
        OpenApiParameter(
            'recursive',
            type=bool,
            description='List recursively (parents, grandparents, etc.)',
            required=False,
        ),
    ]
)
@extend_schema_view(
    get=extend_schema(operation_id='ListElementParents'),
    delete=extend_schema(operation_id='DestroyElementParents', description=(
        'Delete parent elements in bulk.\n\n'
        "Requires an **admin** access to the element's corpus."
    )),
)
class ElementParents(ElementsListBase):
    """
    List all parents of an element.

    Requires a **read** access to the element's corpus.
    """

    @property
    def is_recursive(self):
        recursive_param = self.clean_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    @property
    def selected_corpus(self):
        return self.element.corpus

    def get_queryset(self):
        self.element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        self.check_corpus_access(self.element.corpus)

        return Element.objects.get_ascending(self.kwargs['pk'], recursive=self.is_recursive)


@extend_schema(
    parameters=[
        OpenApiParameter(
            'recursive',
            type=bool,
            description='List recursively (children, grandchildren, etc.)',
            required=False,
        ),
    ]
)
@extend_schema_view(
    # In the frontend, ListElementChildren uses 'position' for ordering by default,
    # to display elements according to their ElementPath.ordering; it makes sense
    # for the API to use the same default ordering as the frontend.
    get=extend_schema(operation_id='ListElementChildren', parameters=[
        OpenApiParameter(
            'order',
            description='Sort elements by a specific field.',
            enum=ELEMENT_ORDERING_FIELDS | {'position'},
            default='position',
            required=False,
        )
    ]),
    delete=extend_schema(operation_id='DestroyElementChildren', description=(
        'Delete child elements in bulk.\n\n'
        "Requires an **admin** access to the element's corpus."
    )),
)
class ElementChildren(ElementsListBase):
    """
    List all children of an element.

    Requires a **read** access to the element's corpus.
    """

    @property
    def is_recursive(self):
        recursive_param = self.clean_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    @property
    def selected_corpus(self):
        return self.element.corpus

    def get_filters(self) -> Q:
        filters = super().get_filters()

        if self.is_recursive:
            filters &= Q(paths__path__overlap=[self.kwargs['pk']])
        else:
            filters &= Q(paths__path__last=self.kwargs['pk'])

        return filters

    def get_order_by(self):
        """
        Override the ElementsListBase ordering to default by path ordering.
        The only endpoint where a path ordering is made visible in the API
        is ListElementNeighbors, in a field named `position`.
        """
        if self.clean_params.get('order', 'position').lower() == 'position':
            direction = self.get_order_direction()
            return (f'{direction}paths__ordering', f'{direction}id')
        return super().get_order_by()

    def get_queryset(self):
        self.element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        self.check_corpus_access(self.element.corpus)

        # Let `self.get_filters` handle filtering optimisations
        if self.clean_params.get('order', 'position').lower() == 'position':
            # This condition is necessary because when ordering by position ('paths__ordering')
            # we run into this bug https://gitlab.com/teklia/arkindex/backend/-/issues/769 and unless
            # the ordering is also used in the .distinct(), this leads to some results being
            # "cut off" from the response. For example, if you have 5 distinct elements, but two
            # of these elements have multiple paths with different orderings within the same parent,
            # the DISTINCT will work as expected during the COUNT query, and find 5 elements;
            # however when next the elements are selected, the path ordering is used in the
            # SELECT query which triggers the "duplicating" bug
            # (https://gitlab.com/teklia/arkindex/backend/-/issues/76).This SELECT would return 7 elements
            # (2 duplicates) but the previous result from the COUNT would mean that only 5 of these
            # 7 elements would get returned, and some elements would therefore (randomly) be missing
            # from the results. Adding the 'paths__ordering' to the .distinct() here forces the
            # duplicating bug to happen during the COUNT query as well, thus avoiding disappearing
            # elements.
            return Element.objects.all().distinct('id', 'paths__ordering')
        return Element.objects.all().distinct()


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(description="Retrieve a single element's information and metadata"),
    patch=extend_schema(description='Rename an element'),
    put=extend_schema(description="Edit an element's attributes. Requires a write access on the corpus."),
    delete=extend_schema(
        description='Delete an element. Requires either an admin access on the corpus, '
                    'or a write access and to be the creator of this element.',
        parameters=[
            OpenApiParameter(
                'delete_children',
                description='Delete all child elements of this element recursively. '
                            'By default, this only removes this element as a parent of the child elements, '
                            'without deleting any child element.',
                type=bool,
                required=False,
            ),
        ],
    ),
)
class ElementRetrieve(ACLMixin, RetrieveUpdateDestroyAPIView):
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_object(self):
        # Prevent duplicating database request
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        queryset = Element.objects.filter(corpus__in=corpora)
        if self.request and self.request.method == 'DELETE':
            # Only include corpus and creator for ACL check and ID for deletion
            return queryset.select_related('corpus').only('id', 'creator_id', 'corpus')

        return queryset \
            .select_related(
                'corpus',
                'type',
                'image__server',
                'creator',
                'worker_run'
            ) \
            .prefetch_related(Prefetch('classifications', queryset=classifications_queryset)) \
            .annotate(metadata_count=Count('metadatas'))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE' and obj.creator_id != request.user.id:
            role = Role.Admin
        if not self.has_access(obj.corpus, role.value):
            access_repr = 'admin' if role == Role.Admin else 'write'
            raise PermissionDenied(detail=f'You do not have {access_repr} access to this element.')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET' and 'pk' in self.kwargs:
            context['element'] = self.get_object()
        return context

    def delete(self, request, *args, **kwargs):
        self.check_object_permissions(self.request, self.get_object())

        queryset = Element.objects.filter(id=self.kwargs['pk'])
        delete_children = self.request.query_params.get('delete_children', 'false').lower() not in ('false', '0')

        element_trash(queryset, user_id=self.request.user.id, delete_children=delete_children)

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementNeighbors',
        tags=['elements'],
        parameters=[
            OpenApiParameter(
                'n',
                type={'type': 'integer', 'minimum': 1, 'maximum': 10},
                description='Number of neighbors to retrieve around the element',
                required=False,
            )
        ],
    )
)
class ElementNeighbors(ACLMixin, ListAPIView):
    """
    List neighboring elements.

    Requires a **read** access to the element's corpus.
    """
    serializer_class = ElementNeighborsSerializer
    # For OpenAPI type discovery
    queryset = Element.objects.none()

    def get_queryset(self):
        n = self.request.query_params.get('n', 1)
        try:
            n = int(n)
        except (TypeError, ValueError):
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})
        if not 1 <= n <= 10:
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})

        element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )

        # Check access permission
        if not self.has_access(element.corpus, Role.Guest.value):
            raise PermissionDenied(detail='You do not have a read access to this element.')

        return Element.objects.get_neighbors(element, n)


@extend_schema(tags=['elements'], request=None)
@extend_schema_view(
    post=extend_schema(operation_id='CreateElementParent', description='Link an element to a new parent'),
    delete=extend_schema(operation_id='DestroyElementParent', description='Delete the relation between an element and one of its parents'),
)
class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RemoveSelectionAutoSchema(AutoSchema):
    """
    AutoSchema subclass that adds a requestBody for the RemoveSelection endpoint.
    This is not doable using only @extend_schema since a DELETE request shouldn't
    have a body according to OpenAPI rules.
    """

    def _get_request_body(self):
        if self.method != 'DELETE':
            return super()._get_request_body()

        # To bypass the no-body rule for DELETE methods, we force the DELETE method to POST.
        self.method = 'POST'
        request_body = super()._get_request_body()

        # The body isn't required but it will be ignored by APIStar, so we must always use body={} when calling RemoveSelection.
        request_body['required'] = False

        # Now that we have retrieved the requestBody schema, we can reset the method to its true nature.
        self.method = 'DELETE'
        return request_body


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(
        operation_id='ListSelection',
        description='List all selected elements',
    )
)
class ManageSelection(SelectionMixin, ListAPIView):
    schema = RemoveSelectionAutoSchema()
    serializer_class = ElementListSerializer
    permission_classes = (IsVerified, )

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        corpus_id = self.request.query_params.get('corpus')
        if corpus_id:
            try:
                UUID(corpus_id)
            except ValueError:
                raise ValidationError({'corpus': 'Not a valid UUID.'})
            get_object_or_404(corpora, pk=corpus_id)
        queryset = self.get_selection(corpus_id=corpus_id)
        return queryset.select_related('type', 'corpus', 'image__server').order_by('corpus', 'type__slug', 'name', 'id')

    # TODO: Use many=True on the response without causing Spectacular to see a paginated list
    @extend_schema(
        operation_id='AddSelection',
        description='Add specific elements',
        responses={201: ElementSlimSerializer},
        request=inline_serializer(
            name="AddSelectionBodySerializer",
            fields={'ids': serializers.ListField(child=serializers.UUIDField())}
        )
    )
    def post(self, request, *args, **kwargs):
        if 'ids' not in request.data:
            raise ValidationError({'ids': ['Element IDs are required.']})

        # Strange field validation done here,
        # as we have different serializers on each request method on the same endpoint.
        # Using serializer fields means we still preserve DRF's normal validation.
        field = serializers.ListField(child=serializers.UUIDField())
        element_ids = field.to_internal_value(request.data['ids'])
        # Silently de-duplicate element ids
        element_ids = set(element_ids)

        elements = list(Element.objects.filter(
            id__in=element_ids,
            corpus__in=Corpus.objects.readable(request.user)
        ))

        if len(elements) != len(element_ids):
            raise ValidationError({'ids': [
                'Some element IDs do not exist or you do not have permission to access them.'
            ]})

        # Add new Selection objects in bulk, ignoring elements that were already selected
        Selection.objects.bulk_create(
            (
                Selection(user=self.request.user, element=element)
                for element in elements
            ),
            ignore_conflicts=True,
        )

        prefetch_related_objects(elements, 'corpus', 'image__server', 'type')
        return Response(
            status=status.HTTP_201_CREATED,
            data=ElementSlimSerializer(
                elements,
                context={'request': request},
                many=True
            ).data,
        )

    @extend_schema(
        operation_id='RemoveSelection',
        description='Remove a specific element or delete any selection',
        request=inline_serializer(
            name="RemoveSelectionBodySerializer",
            fields={'id': serializers.UUIDField(required=False)}
        )
    )
    def delete(self, request, *args, **kwargs):
        if request.data.get('id'):
            field = serializers.PrimaryKeyRelatedField(queryset=self.get_selection())
            element = field.to_internal_value(request.data['id'])
            request.user.selections.get(element=element).delete()
        elif request.data.get('corpus'):
            corpora = Corpus.objects.readable(request.user)
            try:
                UUID(request.data['corpus'])
            except ValueError:
                raise ValidationError({'corpus': 'Not a valid UUID.'})
            corpus = get_object_or_404(corpora, pk=request.data['corpus'])
            request.user.selections.filter(element__corpus=corpus).delete()
        else:
            request.user.selections.all().delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(description='List corpora with their access rights'),
    post=extend_schema(description='Create a new corpus'),
)
class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        # We need to prefetch the access rights on each corpora
        accessible_corpora = filter_rights(self.request.user, Corpus, Role.Guest.value)

        # Keep non duplicated corpora with the max access level
        corpora_level = defaultdict(int)
        for c in accessible_corpora:
            if c.max_level > corpora_level[c.id]:
                corpora_level[c.id] = c.max_level

        corpora = Corpus.objects \
            .filter(id__in=corpora_level) \
            .annotate(authorized_users=Count('memberships')) \
            .prefetch_related('types') \
            .order_by('name', 'id')

        for c in corpora:
            c.access_level = corpora_level[c.id]

            # Avoid making any extra queries to retrieve the top_level_type.
            # We already have prefetched the types, so we can use corpus.types.all()
            # when a corpus has a top_level_type_id to retrieve it.
            if c.top_level_type_id is not None:
                c.top_level_type = next(
                    element_type
                    for element_type in c.types.all()
                    if element_type.id == c.top_level_type_id
                )

        return corpora


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(description=dedent("""
        Retrieve a single corpus.

        Requires a **guest** access for a private corpus, and no specific rights otherwise.
    """).strip()),
    put=extend_schema(description=dedent("""
        Update a corpus.

        Requires an **admin** access on the corpus.
    """).strip()),
    patch=extend_schema(description=dedent("""
        Partially update a corpus.

        Requires an **admin** access on the corpus.
    """).strip()),
    delete=extend_schema(description=dedent("""
        Delete a corpus.

        Requires an **admin** access on the corpus.

        This triggers an asynchronous deletion of a corpus and returns a response immediately;
        the actual deletion will become visible once the job is completed.
        Use the `ListJobs` endpoint to monitor the job's state.
    """).strip())
)
class CorpusRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user) \
            .annotate(authorized_users=Count('memberships')) \
            .prefetch_related('types')

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        # Corpus.objects.readable() already checks that the user has read access
        if self.request.method in permissions.SAFE_METHODS:
            return

        if not self.has_admin_access(obj):
            raise PermissionDenied(detail='You do not have admin access to this corpus.')

    def perform_destroy(self, instance):
        corpus_delete(instance, user_id=self.request.user.id)


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


@extend_schema_view(
    get=extend_schema(
        operation_id='ListTranscriptions',
        tags=['transcriptions'],
        parameters=[
            OpenApiParameter(
                'recursive',
                type=bool,
                description='Recursively list transcriptions on sub-elements',
                required=False,
            ),
            OpenApiParameter(
                'worker_version',
                type=UUID_OR_FALSE,
                description='Only include transcriptions created by a specific worker version. '
                            'If set to `False`, only include transcriptions created by humans.',
                required=False,
            ),
            OpenApiParameter(
                'element_type',
                description='Filter transcriptions by element type',
                required=False,
            )
        ]
    )
)
class ElementTranscriptions(ListAPIView):
    """
    List all transcriptions for an element, optionally filtered by type or worker version id.
    Recursive parameter allow listing transcriptions on sub-elements,
    otherwise element fields in the response will be set to null.
    """
    serializer_class = ElementTranscriptionSerializer
    pagination_class = TranscriptionsPagination
    queryset = Transcription.objects.none()

    @property
    def is_recursive(self):
        if not self.request:
            return
        recursive = self.request.query_params.get('recursive')
        return recursive is not None and recursive.lower() not in ('false', '0')

    @cached_property
    def element(self):
        element = get_object_or_404(Element.objects.filter(
            id=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user)
        ))
        self.check_object_permissions(self.request, element)
        return element

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Do serialize the element attached to each transcription in recursive mode only
        context['ignore_element'] = not self.is_recursive
        return context

    def get_queryset(self):
        # ORDER BY casting IDs as char to avoid the PostgreSQL optimizer's inefficient scan
        # TODO: See if select_related is faster than a prefetch on this endpoint
        queryset = Transcription.objects \
            .prefetch_related('worker_version') \
            .annotate(char_id=Cast('id', output_field=CharField())) \
            .order_by('char_id')

        if self.is_recursive:
            # The transcription's `element` field is only included when recursive=true,
            # so we add the prefetch here
            queryset = queryset.prefetch_related('element__image__server', 'element__type')

            queryset = queryset.filter(element__in=(
                Element
                .objects
                # Transcriptions from the current element
                .filter(id=self.element.id)
                # We are about to use a UNION; we need to explicitly say we only want the ID column in the SELECT,
                # because Django will otherwise pick all Element attributes, which is not supported by `__in`.
                .values('id')
                # Add the current element and the child elements together using a UNION
                # Using Q(element_id=…) | Q(element__in=…) makes PostgreSQL use very slow nested loops and multi-processing.
                .union(
                    # Transcriptions from all children of the current element.
                    # We are not using get_descending, because it includes an ORDER BY clause and a DISTINCT clause,
                    # which are both unnecessary here.
                    ElementPath.objects.filter(path__contains=[self.element.id]).values('element_id'),
                    # Use UNION ALL so that PostgreSQL does not unnecessarily sort and deduplicate UUIDs
                    all=True,
                )
            ))
        else:
            queryset = queryset.filter(element_id=self.element.id)

        return queryset

    def filter_queryset(self, queryset):
        # Filter by worker version
        if 'worker_version' in self.request.query_params:
            worker_version_id = self.request.query_params['worker_version']
            if worker_version_id.lower() in ('false', '0'):
                # Restrict to transcriptions without worker versions
                queryset = queryset.filter(worker_version_id=None)
            else:
                try:
                    queryset = queryset.filter(worker_version_id=worker_version_id)
                except DjangoValidationError as e:
                    raise ValidationError({'worker_version': e.messages})

        # Filter by element_type
        element_type = self.request.query_params.get('element_type')
        if element_type:
            queryset = queryset.select_related('element__type').filter(element__type__slug=element_type)

        return queryset


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElement',
        tags=['elements'],
        parameters=[
            OpenApiParameter(
                'slim_output',
                type=bool,
                description="When enabled, the endpoint will only return the new Element's ID."
                            "This is especially useful for an automated tool, as it speeds up the element creation.",
                required=False,
            )
        ],
        # Return either a full element or the element ID depending on slim_output
        responses=PolymorphicProxySerializer(
            component_name='CreateElementResponse',
            resource_type_field_name='slim_output',
            serializers={
                True: inline_serializer('CreateElementSlimOutput', fields={'id': serializers.UUIDField()}),
                False: ElementSerializer,
            }
        )
    )
)
class ElementsCreate(CreateAPIView):
    """
    Create a new element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer


@extend_schema_view(
    get=extend_schema(
        operation_id='ListElementMetaData',
        description='List all metadata linked to an element.',
        tags=['elements'],
        parameters=[
            OpenApiParameter(
                'load_parents',
                description="List metadata for both the specified element and all of its parents.",
                type=bool,
                default=False,
            )
        ]
    ),
    post=extend_schema(
        operation_id='CreateMetaData',
        tags=['elements'],
    )
)
class ElementMetadata(ListCreateAPIView):
    """
    Create a metadata on an existing element.

    The `date` type allows dates to be parsed and indexed for search APIs. The supported date formats are as follows:

    * `YYYY`: `1995` (recommended)
    * `YYYY-MM`: `1995-02` (recommended)
    * `YYYY-MM-DD`: `1995-02-16` (recommended)
    * `/YYYY(.*)month/`: `1995, february`, with months in French or English (case-insensitive)
    * `YYYY-YYYY`: `1995-1996` (yields a date interval)

    The `reference` type is also indexed in search APIs and allows searching by
    known identifiers or references, or any arbitrary string.

    Values on metadata with a `markdown` type will allow Markdown syntax with a limited set of HTML tags.
    The value will be sanitized and HTML entities will be escaped, but Markdown rendering has to be performed
    on the client side.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    pagination_class = None
    serializer_class = MetaDataCreateSerializer

    def list_metadata(self, element):
        """
        Returns the queryset to list metadata.
        In case the load_parents flag is set, lists metadata for both the specified element and all of its parents
        """
        with_parents = self.request.query_params.get('load_parents', 'false').lower() not in ('false', '0')
        if with_parents:
            qs = MetaData.objects.filter(
                # Filter metadata on the current element and its parents
                element_id__in=(
                    # We use a UNNEST clause to list parents from all paths, avoiding a slow
                    # JOIN with the elements table using `Element.objects.get_ascending`.
                    # https://redmine.teklia.com/projects/arkindex/wiki/Optimizing_parent_elements_metadata
                    element.paths
                    .annotate(parent_id=Unnest('path'))
                    .values('parent_id')
                    .union(
                        Element.objects
                        # Use an annotation to execute a simple SELECT '<uuid>'::uuid FROM <table> LIMIT 1
                        # in order to also fetch element's metadata with no supplementary query.
                        .annotate(element_id=Value(element.id))
                        .values('element_id')[:1]
                    )
                )
            )
        else:
            qs = element.metadatas.all()
        return qs.select_related('entity__worker_run', 'worker_run')

    def get_queryset(self):
        if self.request and self.request.method == 'GET':
            element = get_object_or_404(Element, id=self.kwargs['pk'], corpus__in=Corpus.objects.readable(self.request.user))
            return self.list_metadata(element)

        # We need to use the default database to avoid stale read
        # when a metadata is created immediately after an element creation.
        # Note that this returns a queryset of elements and not metadata!
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.request:
            return context
        if self.request.method != 'GET':
            context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        instance = serializer.save()
        if self.request.user.is_admin or self.request.user.is_internal:
            AllowedMetaData.objects.get_or_create(name=instance.name, type=instance.type, corpus=instance.element.corpus)


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateMetaDataBulk',
        tags=['elements'],
    )
)
class ElementMetadataBulk(CreateAPIView):
    """
    Create multiple metadata on an existing element.
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataBulkSerializer

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when metadata are created immediately after an element creation
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Ignore this step when generating the schema with OpenAPI
        if not self.request:
            return context
        context['element'] = self.get_object()
        return context

    def perform_create(self, serializer):
        instances = serializer.save()
        if self.request.user.is_admin or self.request.user.is_internal:
            AllowedMetaData.objects.bulk_create(
                [
                    AllowedMetaData(name=instance.name, type=instance.type, corpus=instance.element.corpus)
                    for instance in instances['metadata_list']
                ],
                ignore_conflicts=True
            )


@extend_schema(tags=['elements'])
@extend_schema_view(
    get=extend_schema(operation_id='RetrieveMetaData', description='Retrieve an existing metadata'),
    patch=extend_schema(operation_id='PartialUpdateMetaData', description='Partially update an existing metadata'),
    put=extend_schema(operation_id='UpdateMetaData', description='Update an existing metadata'),
    delete=extend_schema(operation_id='DestroyMetaData', description='Delete an existing metadata')
)
class MetadataEdit(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Edit an existing element medadata
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataSerializer

    def get_queryset(self):
        # Filter readable metadata in order to check object permissions
        return (
            MetaData
            .objects
            .select_related('element__corpus', 'worker_run', 'entity__worker_run')
            .filter(element__corpus__in=Corpus.objects.readable(self.request.user))
        )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if not self.has_access(obj.element.corpus, Role.Contributor.value):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        serializer = self.get_serializer(instance)
        serializer.delete(instance)


class AllowedMetaDataPagination(PageNumberPagination):
    page_size = 200


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(operation_id='ListCorpusAllowedMetaDatas'),
    post=extend_schema(
        operation_id='CreateAllowedMetaData',
        description='Create an allowed meta data in a corpus',
    )
)
class CorpusAllowedMetaData(CorpusACLMixin, ListCreateAPIView):
    """
    List (type, name) couples allowed for corpus metadata
    """
    serializer_class = CorpusAllowedMetaDataSerializer
    pagination_class = AllowedMetaDataPagination
    permission_classes = (IsVerifiedOrReadOnly, )
    queryset = AllowedMetaData.objects.none()

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method.lower() == 'post':
            role = Role.Admin
        return self.get_corpus(self.kwargs['pk'], role=role)

    def get_queryset(self):
        return AllowedMetaData.objects.filter(corpus=self.corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus_id'] = self.corpus.id
        return context


@extend_schema(tags=['corpora'])
@extend_schema_view(
    get=extend_schema(operation_id='RetrieveAllowedMetaData', description='Retrieve an allowed metadata.'),
    patch=extend_schema(operation_id='PartialUpdateAllowedMetaData', description='Edit an allowed metadata.'),
    put=extend_schema(operation_id='UpdateAllowedMetaData', description='Edit an allowed metadata.'),
    delete=extend_schema(operation_id='DestroyAllowedMetaData', description='Delete an existing allowed metadata.')
)
class AllowedMetaDataEdit(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    permission_classes = (IsVerified, )
    serializer_class = CorpusAllowedMetaDataSerializer
    queryset = AllowedMetaData.objects.none()

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method.lower() not in permissions.SAFE_METHODS:
            role = Role.Admin
        return self.get_corpus(self.kwargs['corpus'], role=role)

    def get_queryset(self):
        return AllowedMetaData.objects.filter(corpus=self.corpus)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context['corpus_id'] = self.corpus.id
        return context


@extend_schema(tags=['elements'])
class ElementTypeCreate(CreateAPIView):
    """
    Create a new element type on a corpus.

    This requires admin access to the corpus.
    """
    serializer_class = ElementTypeSerializer
    permission_classes = (IsVerifiedOrReadOnly, )


@extend_schema_view(
    get=extend_schema(
        operation_id='RetrieveElementType',
        description=dedent("""
        Retrieve an existing element type.

        This requires read access to the element type's corpus.
        """),
    ),
    put=extend_schema(
        operation_id='UpdateElementType',
        description=dedent("""
        Update an existing element type.

        This requires admin access to the element type's corpus.
        """),
    ),
    patch=extend_schema(
        operation_id='PartialUpdateElementType',
        description=dedent("""
        Update parts of an existing element type.

        This requires admin access to the element type's corpus.
        """),
    ),
    delete=extend_schema(
        operation_id='DestroyElementType',
        description=dedent("""
        Delete an existing element type if it has no associated elements.

        This requires admin access to the element's type corpus.
        """)
    )
)
@extend_schema(tags=['elements'])
class ElementTypeUpdate(RetrieveUpdateDestroyAPIView):
    serializer_class = ElementTypeLightSerializer
    permission_classes = (IsVerifiedOrReadOnly, )

    def get_queryset(self):
        if self.request.method in permissions.SAFE_METHODS:
            corpora = Corpus.objects.readable(self.request.user)
        else:
            corpora = Corpus.objects.admin(self.request.user)

        return ElementType.objects.filter(corpus__in=corpora)

    def perform_destroy(self, instance):
        if instance.elements.exists():
            raise ValidationError({'detail': ['Some elements are using this element type.']})

        # Update Process element type foreign keys manually because letting Django do that can fill up the RAM
        instance.folder_imports.update(folder_type=None)
        instance.element_imports.update(element_type=None)
        super().perform_destroy(instance)


@extend_schema_view(
    post=extend_schema(
        operation_id='CreateElements',
        tags=['elements'],
        responses=inline_serializer(
            'ElementBulkCreateResponse',
            fields={'id': serializers.UUIDField()},
            many=True,
        ),
    )
)
class ElementBulkCreate(CreateAPIView):
    """
    Create multiple child elements at once on a single parent
    """
    serializer_class = ElementBulkSerializer
    permission_classes = (IsVerified, )
    pagination_class = None

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
            if not self.element.image_id:
                raise ValidationError({'element': ['Cannot create child elements on an element without a zone.']})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        return Element \
            .objects \
            .using('default') \
            .filter(corpus__in=Corpus.objects.writable(self.request.user)) \
            .select_related('image') \
            .only('id', 'corpus_id', 'rotation_angle', 'mirrored', 'image')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = self.perform_create(serializer)
        headers = self.get_success_headers({})
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        worker_version = serializer.validated_data.get('worker_version')
        worker_run = serializer.validated_data.get('worker_run')
        elements = serializer.validated_data['elements']

        # ElementPath.get_next_order, but for multiple types
        next_orderings = defaultdict(
            int,
            ElementPath
            .objects
            .using('default')
            .filter(
                # Using __overlap before __last makes Postgres use the GIN index instead of scanning the whole table
                path__overlap=[self.element.id],
                path__last=self.element.id,
                element__type_id__in=set(element_data['type'] for element_data in elements)
            )
            .values('element__type_id')
            .annotate(next_ordering=Max('ordering') + 1)
            .values_list('element__type_id', 'next_ordering')
        )
        # Prepare elements
        for element_data in elements:
            # Keep ordering in the element payload to optimize element path creation later
            element_data['ordering'] = next_orderings[element_data['type']]
            next_orderings[element_data['type']] += 1

            element_data['element'] = Element(
                id=uuid.uuid4(),
                corpus_id=self.element.corpus_id,
                type_id=element_data['type'],
                name=element_data['name'],
                worker_version=worker_version,
                worker_run=worker_run,
                image_id=self.element.image.id,
                polygon=element_data['polygon'],
                rotation_angle=self.element.rotation_angle,
                mirrored=self.element.mirrored,
                confidence=element_data['confidence'],
            )

        Element.objects.bulk_create([
            element_data['element'] for element_data in elements
        ])

        # Support top level elements, by adding an empty initial path to trigger loops below
        parent_paths = list(self.element.paths.all().values_list('path', flat=True))
        if not parent_paths:
            parent_paths = [[]]

        ElementPath.objects.bulk_create([
            ElementPath(
                path=parent_path + [self.element.id],
                element_id=element_data['element'].id,
                ordering=element_data['ordering'],
            )
            for parent_path in parent_paths
            for element_data in elements
        ])

        return [{'id': element_data['element'].id} for element_data in elements]


@extend_schema_view(
    delete=extend_schema(
        operation_id='DestroyCorpusSelection',
        tags=['elements'],
    )
)
class CorpusSelectionDestroy(CorpusACLMixin, SelectionMixin, DestroyAPIView):
    """
    Delete all selected elements on a corpus
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerified, )

    def delete(self, request, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['pk'], role=Role.Admin)
        selected_elements = self.get_selection(corpus_id=corpus.id)
        if not selected_elements.exists():
            raise NotFound

        for batch in range(0, selected_elements.count(), 50):
            queryset = Element.objects.filter(id__in=list(selected_elements[batch:batch + 50].values_list('id', flat=True)))
            element_trash(queryset, user_id=self.request.user.id)

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    delete=extend_schema(
        operation_id='DestroyWorkerResults',
        parameters=[
            OpenApiParameter(
                'worker_version_id',
                type=UUID,
                required=False,
                description='Only delete Worker Results produced by a specific worker version',
            ),
            OpenApiParameter(
                'element_id',
                type=UUID,
                required=False,
                description='Only delete Worker Results under this parent element',
            ),
            OpenApiParameter(
                'use_selection',
                type=bool,
                required=False,
                description='Only delete Worker Results on selected elements in this corpus. '
                            'Cannot be used together with `element_id`.',
            ),
        ],
        tags=['ml'],
    )
)
class WorkerResultsDestroy(CorpusACLMixin, DestroyAPIView):
    """
    Delete all Worker Results from all WorkerVersions or a specific one
    on a Corpus or under a specified parent element (parent element included)
    """
    permission_classes = (IsVerified, )

    # https://github.com/tfranzel/drf-spectacular/issues/308
    @extend_schema(responses={204: None})
    def delete(self, request, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['corpus'], role=Role.Admin)

        errors = defaultdict(list)

        use_selection = self.request.query_params.get('use_selection', 'false').lower() not in ('false', '0')
        if use_selection:
            # Only check for selected elements if the selection feature is enabled
            if settings.ARKINDEX_FEATURES['selection']:
                if 'element_id' in self.request.query_params:
                    errors['use_selection'].append('use_selection and element_id cannot be used simultaneously.')
                if not self.request.user.selected_elements.filter(corpus=corpus).exists():
                    errors['use_selection'].append('No elements of the specified corpus have been selected.')
            else:
                errors['use_selection'].append('Selection is not available on this instance.')

        element_id = None
        if 'element_id' in self.request.query_params:
            try:
                element_id = UUID(self.request.query_params['element_id'])
            except (TypeError, ValueError):
                errors['element_id'].append('Invalid UUID.')
            else:
                if not corpus.elements.filter(id=element_id).exists():
                    errors['element_id'].append('This element does not exist in the specified corpus.')

        worker_version = None
        if 'worker_version_id' in self.request.query_params:
            try:
                worker_version_id = UUID(self.request.query_params['worker_version_id'])
            except (TypeError, ValueError):
                errors['worker_version_id'].append('Invalid UUID.')
            else:
                try:
                    # Includes the worker and revision to get a nice job display name
                    worker_version = WorkerVersion.objects.select_related('worker', 'revision').get(id=worker_version_id)
                except WorkerVersion.DoesNotExist:
                    errors['worker_version_id'].append('This worker version does not exist.')

        if errors:
            raise ValidationError(errors)

        if use_selection:
            selection_worker_results_delete(
                corpus=corpus,
                version=worker_version,
                user_id=self.request.user.id,
            )
        else:
            worker_results_delete(
                corpus_id=corpus.id,
                version=worker_version,
                element_id=element_id,
                user_id=self.request.user.id,
            )

        return Response(status=status.HTTP_204_NO_CONTENT)


@extend_schema_view(
    post=extend_schema(operation_id='MoveElement', tags=['elements']),
)
class ElementMove(CreateAPIView):
    """
    Move an element to a new destination folder
    """
    serializer_class = ElementDestinationSerializer
    permission_classes = (IsVerified, )

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        source = serializer.validated_data['source']
        destination = serializer.validated_data['destination']
        serializer.perform_create_checks(source, destination)

        move_element(source=source, destination=destination, user_id=self.request.user.id)

        return Response(serializer.data, status=status.HTTP_200_OK)


@extend_schema_view(
    post=extend_schema(operation_id='MoveSelection', tags=['elements'])
)
class SelectionMove(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Move selected elements from a given corpus to a new destination folder
    """
    serializer_class = SelectionMoveSerializer
    permission_classes = (IsVerified,)


@extend_schema_view(
    post=extend_schema(operation_id='CreateParentSelection', tags=['elements']),
)
class SelectionParentCreate(CreateAPIView):
    """
    Add a new parent to all selected elements from a given corpus.
    """
    serializer_class = SelectionParentCreateSerializer
    permission_classes = (IsVerified, )
