from django.conf import settings
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_page
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.exceptions import APIException, PermissionDenied
from rest_framework.serializers import CharField, Serializer

from arkindex.documents.models import Corpus
from arkindex.process.models import Process, ProcessMode, Repository, Worker
from arkindex.project.pagination import CustomCursorPagination
from arkindex.training.models import Model
from arkindex.users.models import Role
from arkindex.users.utils import check_level_param, filter_rights, get_max_level


class ACLMixin(object):
    """
    Access control mixin using the generic Right table.
    """
    _user = None

    def __init__(self, *args, user=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._user = user

    @property
    def user(self):
        return self._user or self.request.user

    def has_access(self, instance, level, skip_public=False):
        """
        Check if the user has access to a generic instance with a minimum level
        If skip_public parameter is set to true, exclude rights on public instances
        """
        check_level_param(level)

        # Handle special authentications
        if not skip_public and level <= Role.Guest.value and getattr(instance, 'public', False):
            return True
        if self.user.is_anonymous:
            return False
        elif self.user.is_admin or self.user.is_internal:
            return True

        return instance.memberships.filter(
            Q(
                # Right directly owned by this user
                Q(user=self.user)
                & Q(level__gte=level)
            )
            | Q(
                # Right owned by the group and by the user
                Q(group__memberships__user=self.user)
                & Q(level__gte=level)
                & Q(group__memberships__level__gte=level)
            )
        ).exists()


class RepositoryACLMixin(ACLMixin):

    @property
    def readable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Guest.value).values('id')
        )

    @property
    def executable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Contributor.value).values('id')
        )

    def has_read_access(self, repo):
        return self.has_access(repo, Role.Guest.value)

    def has_execution_access(self, repo):
        return self.has_access(repo, Role.Contributor.value)

    def has_admin_access(self, repo):
        return self.has_access(repo, Role.Admin.value)


class WorkerACLMixin(ACLMixin):
    """
    Access control mixin for workers
    A public worker is considered as executable (i.e. any user has a contributor access)
    """

    @property
    def executable_workers(self):
        return Worker.objects.filter(
            Q(public=True)
            | Q(id__in=filter_rights(self.user, Worker, Role.Contributor.value).values('id'))
            | Q(repository_id__in=filter_rights(self.user, Repository, Role.Contributor.value).values('id'))
        ).distinct()

    def get_max_level(self, worker):
        # Access right on a worker can be defined by a right on its repository
        worker_level = get_max_level(self.user, worker)
        repo_level = get_max_level(self.user, worker.repository)
        return max(
            filter(None, (worker_level, repo_level)),
            default=None
        )

    def has_worker_access(self, worker, level):
        if worker.public and level <= Role.Contributor.value:
            return True
        if self.has_access(worker, level):
            return True
        # User may be granted access to the worker via a right on its repository
        return self.has_access(worker.repository, level)

    def has_read_access(self, worker):
        return self.has_worker_access(worker, Role.Guest.value)

    def has_execution_access(self, worker):
        return self.has_worker_access(worker, Role.Contributor.value)

    def has_admin_access(self, worker):
        return self.has_worker_access(worker, Role.Admin.value)


class CorpusACLMixin(ACLMixin):

    def get_corpus(self, corpus_id, role=Role.Guest):
        """
        Try to retrieve a corpus according to user rights.
        Raises a 404 if the corpus does not exists and a 403 if the permissions are invalid
        """
        corpus = get_object_or_404(Corpus, id=corpus_id)
        if not self.has_access(corpus, role.value):
            raise PermissionDenied(detail=f'You do not have {str(role).lower()} access to this corpus.')
        return corpus

    @property
    def readable_corpora(self):
        return Corpus.objects.filter(
            id__in=filter_rights(self.user, Corpus, Role.Guest.value).values('id')
        )

    @property
    def writable_corpora(self):
        return Corpus.objects.filter(
            id__in=filter_rights(self.user, Corpus, Role.Contributor.value).values('id')
        )

    def has_read_access(self, corpus):
        return self.has_access(corpus, Role.Guest.value)

    def has_write_access(self, corpus):
        return self.has_access(corpus, Role.Contributor.value)

    def has_admin_access(self, corpus):
        return self.has_access(corpus, Role.Admin.value)


class TrainingModelMixin(ACLMixin):
    """
    Access control mixin for machine learning models
    """

    @property
    def readable_models(self):
        return Model.objects.filter(
            id__in=filter_rights(self.user, Model, Role.Guest.value).values('id')
        )

    @property
    def editable_models(self):
        return Model.objects.filter(
            id__in=filter_rights(self.user, Model, Role.Contributor.value).values('id')
        )

    def has_read_access(self, model):
        return self.has_access(model, Role.Guest.value)

    def has_write_access(self, model):
        return self.has_access(model, Role.Contributor.value)

    def has_admin_access(self, model):
        return self.has_access(model, Role.Admin.value)


class ProcessACLMixin(ACLMixin):
    """
    Rights on processes are determined depending on its nature.
    The process creator always has a read access.
    If the process is attached to a project, the project right apply.
    If the process is a version build, the repository right apply.
    """

    @property
    def readable_processes(self):
        # Return a queryset of accessible process
        return Process.objects.filter(
            Q(creator_id=self.user.id)
            | Q(corpus_id__in=filter_rights(self.user, Corpus, Role.Guest.value).values('id'))
            | Q(revision__repo_id__in=filter_rights(self.user, Repository, Role.Guest.value).values('id'))
        ).distinct()

    def process_access_level(self, process):
        # Return the access level on a single process
        access_levels = []
        if process.corpus_id:
            # Use project right. Covers Images, IIIF, PDF, Repository (IIIF), Elements (Workers) and Transkribus process modes
            access_levels.append(get_max_level(self.user, process.corpus))
        elif process.mode == ProcessMode.Repository and process.revision_id:
            # Use repository right in case of a workers docker build
            access_levels.append(get_max_level(self.user, process.revision.repo))
        return max(filter(None, access_levels), default=None)


class SelectionMixin(object):

    def get_selection(self, corpus_id=None):
        assert settings.ARKINDEX_FEATURES['selection'], 'Selection feature is unavailable'
        queryset = self.request.user.selected_elements.all()
        if corpus_id is not None:
            queryset = queryset.filter(corpus_id=corpus_id)
        return queryset


class DeprecatedExceptionSerializer(Serializer):
    detail = CharField()


class ConflictAPIException(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_code = 'conflict'


class DeprecatedAPIException(APIException):
    status_code = 410
    default_detail = 'This endpoint has been deprecated.'
    default_code = 'deprecated'


class DeprecatedMixin(object):
    # Add this mixin to an APIView to make it deprecated.

    serializer_class = DeprecatedExceptionSerializer
    pagination_class = None
    deprecation_message = None

    @classmethod
    def as_view(cls, *args, **kwargs):
        # A custom schema class cannot be assigned on the mixin as it would be linked to the
        # mixin itself, not a view class, so it would not override anything.
        # We therefore have to apply deprecation overrides on the view itself, and
        # we override as_view specifically as this is easier than writing a metaclass.
        extend_schema_view(**{
            method: extend_schema(
                deprecated=True,
                responses={DeprecatedAPIException.status_code: DeprecatedExceptionSerializer},
            )
            # Apply extend_schema only on methods that exist on this endpoint
            for method in cls.http_method_names
            if hasattr(cls, method)
        })(cls)
        return super().as_view(*args, **kwargs)

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        method = request.method.lower()
        if method in self.http_method_names and hasattr(self, method):
            raise DeprecatedAPIException(detail=getattr(self, 'deprecation_message', None))


class CachedViewMixin(object):
    """
    Add this mixin to any class-based view to cache it.
    """
    cache_timeout = None
    cache_prefix = None

    def __init__(self, *args, **kwargs):
        if self.cache_timeout:
            self.dispatch = cache_page(
                self.cache_timeout,
                key_prefix=self.cache_prefix,
            )(self.dispatch)


class CustomPaginationViewMixin(object):
    """
    A custom cursor pagination mixin
    Elements count can be retrieved with the `with_count` parameter
    """
    pagination_class = CustomCursorPagination

    @property
    def paginator(self):
        if not hasattr(self, '_paginator'):
            params = self.request.query_params
            with_count = params.get('with_count') not in (None, '', 'false', '0')
            self._paginator = self.pagination_class(with_count=with_count)
        return self._paginator
